<?php

class iaPaypal extends abstractCore
{
	const API_VERSION = '109.0';

	const ENDPOINT_LIVE = 'https://api-3t.paypal.com/nvp';
	const ENDPOINT_SANDBOX = 'https://api-3t.sandbox.paypal.com/nvp';

	const URL_LIVE = 'https://www.paypal.com/cgi-bin/webscr';
	const URL_SANDBOX = 'https://www.sandbox.paypal.com/cgi-bin/webscr';

	const PAYMENT_TYPE_SALE = 'Sale';
	const PAYMENT_OPTION_RECURRING = 'RecurringPayments';

	const IPN_RESPONSE_VERIFIED = 'VERIFIED';

	const RESPONSE_SUCCESS = 'SUCCESS';
	const RESPONSE_SUCCESSWITHWARNING = 'SUCCESSWITHWARNING';

	const ITEM_CATEGORY_DIGITAL = 'Digital';
	const ITEM_CATEGORY_PHYSICAL = 'Physical';

	const METHOD_SET_EC = 'SetExpressCheckout';
	const METHOD_DO_EC = 'DoExpressCheckoutPayment';
	const METHOD_GET_EC_DETAILS = 'GetExpressCheckoutDetails';
	const METHOD_CREATE_RECURRING_PAYMENTS_PROFILE = 'CreateRecurringPaymentsProfile';

	// not used
	const IPN_TXN_TYPE_SUBSCR_CANCEL = 'subscr_cancel'; // Subscription canceled
	const IPN_TXN_TYPE_SUBSCR_EOT = 'subscr_eot'; // Subscription expired
	const IPN_TXN_TYPE_SUBSCR_FAILED = 'subscr_failed'; // Subscription payment failed
	const IPN_TXN_TYPE_SUBSCR_MODIFIED = 'subscr_modify'; // Subscription modified
	const IPN_TXN_TYPE_SUBSCR_PAYMENT = 'subscr_payment'; // Subscription payment received
	const IPN_TXN_TYPE_SUBSCR_SIGNUP = 'subscr_signup'; // Subscription started
	//

	protected $_pluginName = 'paypal';

	protected $_token;
	protected $_response;

	// preset of sandbox credentials
	protected $_user = 'sdk-three_api1.sdk.com';
	protected $_password = 'QFZCWN5HZM8VBG7Q';
	protected $_signature = 'A-IzJhZZjhg29XQ2qnhapuwxIDzyAZQ92FRP5dqBzVesOkzbdUONzmOU';

	protected $_endpointUrl = self::ENDPOINT_SANDBOX;
	protected $_url = self::URL_SANDBOX;

	protected $_currencyCode = '';


	public function init()
	{
		parent::init();

		if (!$this->iaCore->get('paypal_demo_mode'))
		{
			$this->_endpointUrl = self::ENDPOINT_LIVE;
			$this->_url = self::URL_LIVE;

			$this->_user = $this->iaCore->get('paypal_api_user');
			$this->_password = $this->iaCore->get('paypal_api_password');
			$this->_signature = $this->iaCore->get('paypal_api_signature');
		}

		$this->_currencyCode = $this->iaCore->get('paypal_currency_code');
	}

	public function getPluginName()
	{
		return $this->_pluginName;
	}

	public function redirect()
	{
		$url = $this->getUrl('cmd=_express-checkout&token=') . urlencode($this->_token);

		header('Location: ' . $url);
		exit;
	}

	public function getUrl($params = '')
	{
		return $this->_url . ($params ? '?' . $params : '');
	}

	public function getIpnUrl()
	{
		$iaPage = $this->iaCore->factory('page', iaCore::FRONT);

		return $iaPage->getUrlByName('ipn_paypal');
	}

	public function getError()
	{
		$code = urldecode($this->getResponse()->l_errorcode0);
		$message = urldecode($this->getResponse()->l_longmessage0);

		return 'Error ' . $code . ': ' . $message;
	}

	public function getToken()
	{
		return $this->getResponse()->token;
	}

	public function getResponse($asPlainArray = false)
	{
		if ($this->_response)
		{
			return $asPlainArray
				? $this->_response
				: (object)$this->_response;
		}

		return null;
	}

	protected function _apiCall($methodName, array $params = array())
	{
		$defaults = array_merge($params, array(
			'method' => $methodName,
			'version' => self::API_VERSION,
			'pwd' => $this->_password,
			'user' => $this->_user,
			'signature' => $this->_signature
		));

		$params = array_merge($defaults, $params);
		$params = http_build_query(array_change_key_case($params, CASE_UPPER));

		$ch = curl_init();

		curl_setopt($ch, CURLOPT_URL, $this->_endpointUrl);
		curl_setopt($ch, CURLOPT_VERBOSE, true);
		curl_setopt($ch, CURLOPT_SSL_VERIFYPEER, false);
		curl_setopt($ch, CURLOPT_SSL_VERIFYHOST, false);
		curl_setopt($ch, CURLOPT_RETURNTRANSFER, true);
		curl_setopt($ch, CURLOPT_POST, true);
		curl_setopt($ch, CURLOPT_POSTFIELDS, $params);

		$response = curl_exec($ch);

		if (curl_errno($ch))
		{
			$result = false;
		}
		else
		{
			$this->_response = $this->_explodeNVP($response);

			$ack = strtoupper($this->getResponse()->ack);
			$result = (self::RESPONSE_SUCCESS == $ack || self::RESPONSE_SUCCESSWITHWARNING == $ack);
		}

		curl_close($ch);

		if ($result && isset($this->getResponse()->token))
		{
			$this->_token = $this->getResponse()->token;
		}

		return $response;
	}

	protected function _explodeNVP($nvp)
	{
		$intial = 0;
		$result = array();

		while (strlen($nvp))
		{
			$keypos = strpos($nvp,'=');
			$valuepos = strpos($nvp,'&') ? strpos($nvp,'&') : strlen($nvp);

			$keyval = substr($nvp, $intial, $keypos);
			$valval = substr($nvp, $keypos+1, $valuepos-$keypos-1);

			$result[strtolower(urldecode($keyval))] = urldecode($valval);
			$nvp = substr($nvp, $valuepos+1, strlen($nvp));
		}

		return $result;
	}

	protected function _getRawPostString()
	{
		$rawPost = explode('&', file_get_contents('php://input'));
		$post = array();

		foreach ($rawPost as $chunk)
		{
			$keyval = explode ('=', $chunk);
			if (count($keyval) == 2)
				$post[$keyval[0]] = urldecode($keyval[1]);
		}

		if (function_exists('get_magic_quotes_gpc') && get_magic_quotes_gpc())
		{
			array_map('stripslashes', $post);
		}

		$result = '';
		foreach ($post as $key => $value)
		{
			$result.= '&' . $key . '=' . urlencode($value);
		}

		return $result;
	}

	public function checkIpnMessage()
	{
		if (!($ch = curl_init($this->getUrl())))
		{
			return false;
		}

		$postData = 'cmd=_notify-validate' . $this->_getRawPostString();

		curl_setopt($ch, CURLOPT_HTTP_VERSION, CURL_HTTP_VERSION_1_1);
		curl_setopt($ch, CURLOPT_POST, true);
		curl_setopt($ch, CURLOPT_RETURNTRANSFER, true);
		curl_setopt($ch, CURLOPT_POSTFIELDS, $postData);
		curl_setopt($ch, CURLOPT_SSL_VERIFYPEER, true);
		curl_setopt($ch, CURLOPT_SSL_VERIFYHOST, 2);
		curl_setopt($ch, CURLOPT_FORBID_REUSE, true);
		curl_setopt($ch, CURLOPT_CONNECTTIMEOUT, 30);
		curl_setopt($ch, CURLOPT_HTTPHEADER, array('Connection: Close'));

		$response = curl_exec($ch);

		if (curl_errno($ch))
		{
			curl_close($ch);

			return false;
		}
		else
		{
			curl_close($ch);
		}

		return (0 == strcmp($response, self::IPN_RESPONSE_VERIFIED));
	}


	/* API CALLS */
	public function setExpressCheckout(array $planInfo, $description, $returnURL, $cancelURL)
	{
		$params = array(
			'brandname' => $this->iaCore->get('site'),
			'localecode' => $this->iaCore->iaView->language,
			'returnurl' => $returnURL,
			'cancelurl' => $cancelURL,
			'allownote' => 1,
			'paymentrequest_0_amt' => (float)$planInfo['cost'],
			'paymentrequest_0_paymentaction' => self::PAYMENT_TYPE_SALE,
			'paymentrequest_0_currencycode' => $this->_currencyCode,
			'paymentrequest_0_desc' => $description
		);

		if (iaUsers::hasIdentity())
		{
			$userInfo = iaUsers::getIdentity(true);

			$params['email'] = $userInfo['email'];
			//$params['buyerid'] = $userInfo['id'];
			//$params['buyerusername'] = $userInfo['fullname'];
		}

		if ($planInfo['recurring'])
		{
			$params['l_billingtype0'] = self::PAYMENT_OPTION_RECURRING;
			$params['l_billingagreementdescription0'] = $description;
		}

		return $this->_apiCall(self::METHOD_SET_EC, $params);
	}

	public function doExpressCheckoutPayment($token, $payer, $amount)
	{
		return $this->_apiCall(self::METHOD_DO_EC, array(
			'token' => $token,
			'payerid' => $payer,
			'paymentrequest_0_amt' => $amount,
			'paymentrequest_0_paymentaction' => self::PAYMENT_TYPE_SALE,
			'paymentrequest_0_currencycode' => $this->_currencyCode,
			//'paymentrequest_0_custom' => (int)iaUsers::getIdentity()->id,
			'notifyurl' => $this->getIpnUrl()
		));
	}

	public function getExpressCheckoutDetails($token)
	{
		return $this->_apiCall(self::METHOD_GET_EC_DETAILS, array('token' => $token));
	}

	public function createRecurringPaymentsProfile(array $planInfo, $description, $token, $payer)
	{
		$params = array(
			'token' => $token,
			'payerid' => $payer,
			'profilestartdate' => gmdate('Y-m-d\TH:i:s\Z'),
			'desc' => $description,
			'billingperiod' => ucfirst($planInfo['unit']),
			'billingfrequency' => $planInfo['duration'],
			'totalbillingcycles' => $planInfo['cycles'],
			'amt' => $planInfo['cost'],
			'currencycode' => $this->_currencyCode
		);

		if (iaUsers::hasIdentity())
		{
			$params['email'] = iaUsers::getIdentity()->email;
			$params['firstname'] = iaUsers::getIdentity()->fullname;
		}

		return $this->_apiCall(self::METHOD_CREATE_RECURRING_PAYMENTS_PROFILE, $params);
	}
	//


	public function handleIpn(array $transactionInfo, array $data)
	{
		$iaSubscription = $this->iaCore->factory('subscription');

		switch ($data['txn_type'])
		{
			case 'recurring_payment_profile_created':
				$values = array(
					'date_created' => date(iaDb::DATETIME_FORMAT, strtotime($data['time_created'])),
					'date_next_payment' => date(iaDb::DATETIME_FORMAT, strtotime($data['next_payment_date'])),
					'status' => iaSubscription::ACTIVE
				);

				$iaSubscription->update($values, $data['recurring_payment_id']);

				break;

			case 'recurring_payment_suspended':
			case 'recurring_payment_suspended_due_to_max_failed_payment':
				$iaSubscription->update(array('status' => iaSubscription::SUSPENDED), $data['recurring_payment_id']);

				break;

			case 'recurring_payment_profile_cancel':
				$iaSubscription->update(array('status' => iaSubscription::CANCELED), $data['recurring_payment_id']);
		}

		if (in_array($data['txn_type'], array('express_checkout', 'recurring_payment', 'send_money')))
		{
			$this->_processIpnPayment($transactionInfo, $data);
		}
	}

	protected function _processIpnPayment(array $transactionInfo, array $data)
	{
		$checks = array(
			'receiver_email' => $this->iaCore->get('paypal_email'),
			'mc_currency' => $this->iaCore->get('paypal_currency_code')
		);

		$error = false;
		foreach ($checks as $key => $expectedValue)
		{
			if (empty($data[$key]) || $data[$key] != $expectedValue)
			{
				$error = true;
				break;
			}
		}

		if (!empty($data['payment_gross']))
		{
			$amount = $data['payment_gross'];
		}
		elseif (!empty($data['mc_gross']))
		{
			$amount = $data['mc_gross'];
		}
		else
		{
			$amount = $data['amount3'];
		}

		if ($amount < $transactionInfo['amount'])
		{
			$error = true;
		}

		if (!$error)
		{
			$values = array();

			$values['email'] = $data['payer_email'];
			$values['subscription_id'] = empty($data['recurring_payment_id']) ? $data['subscr_id'] : $data['recurring_payment_id'];
			$values['reference_id'] = $data['txn_id'];
			$values['amount'] = $amount;
			$values['fullname'] = $data['first_name'] . ' ' . $data['last_name'];

			$iaTransaction = $this->iaCore->factory('transaction');

			switch ($data['payment_status'])
			{
				case 'Completed':
				case 'Canceled_Reversal':
					$values['status'] = iaTransaction::PASSED;
					break;
				case 'Pending':
					$values['status'] = iaTransaction::PENDING;
					break;
				case 'Expired':
				case 'Failed':
				case 'Denied':
					$values['status'] = iaTransaction::FAILED;
					break;
				case 'Refunded':
				case 'Reversed':
					$transaction['status'] = iaTransaction::REFUNDED;
			}

			$iaTransaction->update($values, $transactionInfo['id']);
		}
	}
}